!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2015, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
module mod_fij
use mod_global, only : dp
implicit none
save
private
real(kind=dp),dimension(:,:)  ,allocatable  :: fij

public fij,             &
       read_fij,        &
       process_dedr_fij,&
       deallocate_fij,  &
       allocate_fij

contains

    subroutine allocate_fij
    use mod_neigh, only  : list_length
    use mod_box, only    : dimn
    implicit none

    allocate( fij(dimn,list_length) )
    fij = 0._dp
    end subroutine allocate_fij

    subroutine process_dedr_fij(dedr,at_l,at_m)
    use mod_neigh, only     : list,marker1,marker2
    use mod_box, only : dimn,samplesize,pbc
    use mod_particles, only : at,at_i
    implicit none

    ! Passed variables
    real(kind=dp),intent(in) :: dedr
    integer,      intent(in) :: at_l,at_m

    ! Local variables
    integer j,k
    real(kind=dp) :: rij(dimn),rij_initial(dimn),distij,distij_initial


    do j=marker1(at_l),marker2(at_l)
        k = list(j)
        if (at_m .eq. k) then
            rij = at(at_l)%x-at(at_m)%x
            rij_initial = at_i(at_l)%x-at_i(at_m)%x
            ! ".and. pbc" in the following line is redundant. pbc has alread been
            ! taken care of while forming neighbor lists.
            where(abs(rij) > 0.5_dp * samplesize .and. pbc)
                rij = rij - sign(samplesize,rij)
            end where
            where(abs(rij_initial) > 0.5_dp * samplesize .and. pbc)
                rij_initial = rij_initial - sign(samplesize,rij_initial)
            end where
            distij = dot_product(rij,rij)
            distij_initial = dot_product(rij_initial,rij_initial)
            fij(:,j) = fij(:,j) - dedr * (sqrt(distij)-sqrt(distij_initial)) * rij/sqrt(distij)
            exit
        endif
    enddo
    return
    end subroutine process_dedr_fij

    subroutine deallocate_fij
    implicit none
    if (allocated(fij)     ) deallocate(     fij    )
    end subroutine deallocate_fij

    subroutine read_fij(rcutoff)
    use mod_particles, only : at
    use mod_box, only       : dimn,samplesize,pbc,nat
    use mod_neigh, only     : max_pairs_per_atom,list,list_length,marker1,marker2
    use mod_file, only      : FileExists,iofile
    use mod_output, only    : output, nerror,error,erroroutput
    implicit none

    real(kind=dp), intent(out) :: rcutoff

    integer iprev,inew,ind,fijunit,i,j,k,ierror,ct
    real(kind=dp) :: val(dimn),rij(dimn),distsq
    character(len=80) :: fijfile
    logical :: counted
    integer :: dummy(2,728090)

    real(kind=dp) :: force(dimn,nat)
    !integer :: dummy(2,181509)

    allocate( marker1(nat) )
    allocate( marker2(nat) )
    allocate( list(max_pairs_per_atom * nat) )
    iprev = 0
    ind = 0

    fijfile = 'fij.data'
    if (FileExists(fijfile,.false.)) then
        call iofile(fijfile,'formatted  ',fijunit,.false.,.false.)
    else
        nerror = 2
        error(1) = 'fij information file not found'
        error(2) = 'in subroutine read_fij,  Filename = '&
                    // trim(fijfile)
        call erroroutput
        stop
     endif

    ct = 0
    rcutoff = 0._dp
    do
        read(fijunit,*,iostat=ierror) i,j
        if(ierror .ne. 0) then
            write(*,*) "Reached end of file"
            exit
        endif
        ! Calculate cutoff raidus
                rij = at(i)%x-at(j)%x
                !-----------------------------------------------------------------------------------------
                !if(abs(rij(1)) > 0.5_dp * samplesize(1) .or. &
                !   abs(rij(2)) > 0.5_dp * samplesize(2) .or. &
                !   abs(rij(3)) > 0.5_dp * samplesize(3)       ) write(*,*) i,j,sqrt(dot_product(rij,rij))
                !
                ! ".and. pbc" in the following line is redundant. pbc has alread been
                ! taken care of while forming neighbor lists.
                where(abs(rij) > 0.5_dp * samplesize .and. pbc)
                    rij = rij - sign(samplesize,rij)
                end where
                distsq = dot_product(rij,rij)
                if (distsq .gt. rcutoff) rcutoff = distsq

        ct = ct + 1
        dummy(:,ct) = (/i,j/)
    enddo
    rcutoff = sqrt(rcutoff)
    write(*,*) "Cutoff radius = ", rcutoff

    list_length = ct
    write(*,*) "length = ",list_length

    do ind = 1,list_length
        i = dummy(1,ind)
        inew = i
        j = dummy(2,ind)
        list(ind) = j
        if (inew .ne. iprev) then
            marker1(inew) = ind
            if (inew .ne. 1) marker2(iprev) = ind-1
        endif
        iprev = i
    enddo
    marker2(iprev) = list_length
        
    do i=1,nat
        if (marker1(i) .eq. 0) then
            if (marker2(i) .ne. 0) then
                write(*,*) "something is again wrong!"
                stop
            endif
            marker2(i) = -1
        endif
    enddo


    allocate(fij(dimn,list_length))
    close(fijunit)

    ! Read fij values
    open(unit=fijunit,file='fij.data')
    ind = 0
    do
        read(fijunit,*,iostat=ierror) i,j,val
        if(ierror .ne. 0) then
            write(*,*) "Reached end of file"
            exit
        endif

        counted = .false.
        do k=marker1(i),marker2(i)
            if (j .eq. list(k)) then
                !-------------------------------------------------------------------------------
                ! rij = at(i)%x-at(j)%x
                ! ! ".and. pbc" in the following line is redundant. pbc has alread been
                ! ! taken care of while forming neighbor lists.
                ! where(abs(rij) > 0.5_dp * samplesize .and. pbc)
                !     rij = rij - sign(samplesize,rij)
                ! end where
                ! distij = dot_product(rij,rij)
                ! fij(:,k) = val * rij/sqrt(distij)
                !
                fij(:,k) = val
                counted = .true.
                exit
            endif
        enddo
        if (counted .neqv. .true.) then
            write(*,*) "something wrong",i,j
            stop
        endif
    enddo

    ! Checking forces

    force = 0._dp
    do i=1,nat
        do j=marker1(i),marker2(i)
            k = list(j)
            force(:,i) = force(:,i) + fij(:,j)
            force(:,k) = force(:,k) - fij(:,j)
            !write(*,*) fij(:,j),force(:,1)
        enddo
        !stop
    enddo
    !-------------------------------------------------------------------------------
    ! do i=1,nat
    !     write(*,*) force(:,i)
    ! enddo
    ! stop
    !
     end subroutine read_fij
end module mod_fij
