!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2015, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!****************************************************************************
!**
!**   MODULE mod_forces: 
!**
!**   type definitions:
!**   ----------------
!**   N/A
!**
!**   Real Variable Definitions:
!**   -------------------------------------
!**   fij(:)  - interatomic forces
!**   pfij(:) - interatomic forces
!**   scl
!**
!**   Integer Variable Definitions:
!**   ----------------------------
!**   max_pairs_per_atom - max number of interactions per atom
!**   list_length        - number of interactions
!**   marker1,marker2    -
!**   list
!**  
!**   logical Variable Definitions:
!**   ----------------------------
!**
!**   Contains Routines:
!**   ------------------
!**   ngh_list(range) - Form the neighbor list, including all atom pairs 
!**                     that are within a distance "range" to each other.
!**                     "range", in our case is equal to the interaction 
!**                     range.
!**   compute_forces - 
!****************************************************************************
#include "KIM_API_status.h"
module mod_forces
use, intrinsic :: iso_c_binding
use KIM_API_F03
use mod_global, only : dp
implicit none
save
private
public compute_forces,          &
       set_forces,              &
       deallocate_forces,       &
       rcutoff

real(kind=dp)     :: energy
real(kind=dp)     :: rcutoff
character(len=KIM_KEY_STRING_LENGTH) :: modelname
!character(len=80) :: modelname
real(kind=dp),dimension(:,:),allocatable :: fi


contains

    subroutine set_forces
    !-------------------------------------------------------------------------------
    ! SUBROUTINE set_interatomic :
    !
    ! Passed Parameters :-
    ! -----------------
    !
    ! Non-Obvious Parameter :-
    ! ---------------------
    !
    use mod_global, only : cmdunit
    use mod_output, only : string,output,erroroutput,nerror,error,heading
    use mod_qclib, only : next,downcase,freein
    implicit none
    !-------------------------------------------------------------------------------
    ! local variables
    !
    character(len=80) :: input
    character(len=4) :: cc
    integer :: lower,upper,i,ldif
    logical :: flag,terminate
    !-------------------------------------------------------------------------------
    ! allow indentation with spaces in input file
    !
    heading=.true.
    write(string,'(a)')'**Reading interatomic data**'; call output
    terminate=.false.
    do
        read(cmdunit,'(a80)') input
        i=1
        do while ((input(i:i).eq.' ').and.(i.le.80))
            i=i+1
        enddo
        upper = i-1
        
        lower = upper
        upper = next(lower,input)
        ldif = upper - lower - 1
        if (ldif.eq.0) then
            cc = '    '
        else
            cc = input(lower+1:upper-1)
        end if
        
        lower=upper
        upper=next(lower,input)
        do i=lower,upper
             if ((input(i:i).eq.'T').or.(input(i:i).eq.'t')) then
                flag=.true.
                exit
             else if ((input(i:i).eq.'F').or.(input(i:i).eq.'f')) then
                flag=.false.
                exit
             endif
        enddo
        select case(downcase(cc)) 
        case('modl')
            modelname= trim(input(lower+1:upper-1))
        case('end')
            terminate=.true.
        case default
            if (cc .ne.'    ' .and. cc .ne.'%') then
                nerror = 1
                error(1) = 'macro "' // cc // '" not recognized**'
                call erroroutput
                stop
            endif
        end select

        if (terminate) exit
    enddo

    end subroutine set_forces

    subroutine compute_forces
    !-------------------------------------------------------------------------------
    ! SUBROUTINE compute_forces :
    !     Calculates forces on the atoms by invoking the kim_api
    !     Calculates interatomic forces through standard, projection
    !
    ! Passed Parameters :-
    ! -----------------
    !
    ! Non-Obvious Parameter :-
    ! ---------------------
    !
    use mod_particles, only     : at
    use mod_box, only           : nat,         &
                                  nsp,         &
                                  dimn,        &
                                  samplesize

    use mod_neigh, only         : form_neigh_list

    use mod_fij, only           : allocate_fij


    use mod_rigidity, only      : project,     &
                                  project_forces
    ! Nikhil
    use mod_file, only : FileExists,iofile

    use mod_output, only        : string,      &
                                  output

    use mod_kim
    use KIM_API_F03
    implicit none


    ! KIM variables
  character(len=KIM_KEY_STRING_LENGTH) :: testkimfile="descriptor.kim"
  type(c_funptr)     :: pdedr,pget_neigh


    write(string,'(a)')'Starting force calculation : '; call output
    allocate(fi(dimn,nat))

    !-------------------------------------------------------------------------------
    ! Setup the KIM API object
    ! 1. Broadcast the
    !     a. name of the test and model
    !     b. nature of neigh list
    !     c. dimension
    !     d. number of atoms
    !     e. position of atoms
    !     f. number of species
    !     g. samplesize.
    ! 2. Get the cutoff radius from the model
    !

    ! Calculate Forces and interatomic forces using model via KIM_API

    call setup_KIM_API_object(testkimfile, modelname, &
                              dimn,nat,nsp,at,energy,fi,rcutoff,samplesize)
                          
    write(string,5009) rcutoff; call output
    !-------------------------------------------------------------------------------
    ! Neighbor list calculation for the given cutoff
    !
    call form_neigh_list(rcutoff)

    !-------------------------------------------------------------------------------
    ! Allocate interatomic forces fij
    !
    call allocate_fij

    !-------------------------------------------------------------------------------
    ! Inform KIM where it can find the neighbor list and the get_neigh function
    !
    pdedr= c_funloc(process_dedr)
    pget_neigh = c_funloc(get_neigh)
    call setup_kim_access(pdedr,pget_neigh)

    call model_compute


    !-------------------------------------------------------------------------------
    ! Free memory
    !
    call free_KIM_API_object

    !-------------------------------------------------------------------------------
    ! Compute proj_fij using rigidity theory
    !
    if (project) call project_forces

    5009 format('Cutoff radius from model = ',f5.3)
    end subroutine compute_forces

    integer(c_int) function get_neigh(pkim,mode,request,atom,numnei,pnei1atom,pRij) bind(c)
    use KIM_API_F03
    use mod_neigh, only : marker1,marker2,list
    implicit none

    !-- Transferred variables
    type(c_ptr),    intent(in)  :: pkim
    integer(c_int), intent(in)  :: mode
    integer(c_int), intent(in)  :: request
    integer(c_int), intent(out) :: atom
    integer(c_int), intent(out) :: numnei
    type(c_ptr),    intent(out)  :: pnei1atom
    type(c_ptr),    intent(out)  :: pRij
    integer(c_int), pointer  :: nei1atom(:)

    !-- Local variables
    integer(c_int), pointer :: N; type(c_ptr) :: pnAtoms
    integer(c_int)  mrk1,mrk2,ier,idum
    integer(c_int), save :: current_atom = 0

    pnAtoms = kim_api_get_data(pkim, "numberOfParticles", ier)
    if (ier.lt.KIM_STATUS_OK) then
        idum = kim_api_report_error(__LINE__, __FILE__, "kim_api_get_data", ier)
        stop
    endif
    call c_f_pointer(pnAtoms,N)

    ! check mode and request
    if (mode.eq.0) then ! iterator mode
        ! Iterator mode

        if (request==0) then
            ! Restart iterator
            current_atom = 0      ! restart from first atom next time
            numnei = 0            ! get_neigh is called
            get_neigh = KIM_STATUS_NEIGH_ITER_INIT_OK ! iterator has been successfully initialized
            return
        elseif (request==1) then
            current_atom = current_atom + 1      ! proceed to next atom
            if (current_atom.eq.N) then
                get_neigh = KIM_STATUS_NEIGH_ITER_PAST_END ! error: iterator has been
                                                         ! incremented past end of list
                return
            endif
            atom = current_atom
            numnei = marker2(current_atom) - marker1(current_atom) + 1
            pnei1atom = c_loc(list(marker1(current_atom)))
            call c_f_pointer(pnei1atom,nei1atom,[numnei])
            pRij = c_null_ptr
            get_neigh = KIM_STATUS_OK ! succesful completion
            return
       endif


    elseif (mode.eq.1) then ! locator mode
       if ( (request.gt.N) .or. (request.lt.1)) then
           idum = kim_api_report_error(__LINE__, __FILE__, "Invalid atom ID in get_neigh", KIM_STATUS_PARTICLE_INVALID_ID)
           get_neigh = KIM_STATUS_PARTICLE_INVALID_ID
           return
       endif

       ! set the returned atom
       atom = request
       if (request==N) then
          numnei = 0
          pnei1atom = c_null_ptr
       else
           ! set the returned number of neighbors for the returned atom
            mrk1 = marker1(atom)
            mrk2 = marker2(atom)
            numnei = mrk2-mrk1+1

            ! set the location for the returned neighbor list
            pnei1atom = c_loc(list(mrk1))
            call c_f_pointer(pnei1atom,nei1atom,[numnei])
        endif

        ! set pointer to Rij to NULL
        pRij = c_null_ptr
        get_neigh = KIM_STATUS_OK
        return

    else ! not iterator or locator mode
       idum = kim_api_report_error(__LINE__, __FILE__, "Invalid mode in get_neigh", KIM_STATUS_NEIGH_INVALID_MODE)
       get_neigh = KIM_STATUS_NEIGH_INVALID_MODE
       return
    endif

    get_neigh = KIM_STATUS_FAIL  !  should not get to this line (added to prevent pedantic warning)

    end function get_neigh

    integer(c_int) function process_dedr(pkim,deidr,r,dx,at_i,at_j)bind(c)
    use KIM_API_F03
    use mod_fij,       only : process_dedr_fij
    implicit none

    ! Passed variables
    type(c_ptr),   intent(in) :: pkim,dx
    real(c_double),intent(in) :: deidr,r
    integer(c_int),intent(in) :: at_i,at_j

    call process_dedr_fij(deidr,at_i,at_j)
    process_dedr = KIM_STATUS_OK
    end function process_dedr


    subroutine deallocate_forces
    use mod_fij, only : deallocate_fij
    implicit none
    if (allocated(      fi)) deallocate(        fi  )
    call deallocate_fij
    end subroutine deallocate_forces

end module mod_forces
