!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2015, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
module mod_neigh
use mod_global, only : dp
implicit none
save
private
public marker1,marker2,         &
       list,list_length,        &
       max_pairs_per_atom,      &
       ngh_list,                &
       form_neigh_list,         &
       deallocate_neigh

! Neighbor list for calculation of forces
integer,dimension(:),allocatable :: marker1,marker2
integer,allocatable,target :: list(:)
integer                          :: list_length
integer,parameter                :: max_pairs_per_atom = 500

contains
    !-------------------------------------------------------------------------------
    ! Allocate memory for neighbor lists used for force calculation
    !
    subroutine form_neigh_list(cutoff)
    use mod_box, only : nat
    use mod_output, only        : string,      &
                                  output
    implicit none

    ! Passed varriables
    real(kind=dp), intent(in)   :: cutoff

    integer :: max_list_length

    max_list_length = max_pairs_per_atom * nat
    allocate( marker1(nat) )
    allocate( marker2(nat) )
    allocate( list(max_list_length) )
    call ngh_list(list,list_length,marker1,marker2,cutoff)
    write(string,5010) list_length; call output
    5010 format('Allocating vector fij of size list_length = ',I10)
    end subroutine form_neigh_list

    subroutine ngh_list(lst,lst_lth,mrk1,mrk2,cutoff)
    !-------------------------------------------------------------------------------
    ! SUBROUTINE ngh_list :
    !
    ! Passed Parameters :-
    ! -----------------
    ! lst (out)
    ! lst_lth (out)
    ! mrk1,mrk2 (out)
    !
    ! Non-Obvious Parameter :-
    ! ---------------------
    !
    use mod_particles, only : at
    use mod_box, only : samplesize,pbc,nat,dimn
    use mod_output, only : string,output,erroroutput
    implicit none
    !-------------------------------------------------------------------------------
    ! Transferred Variables
    !
    integer, dimension(:), intent(inout) :: mrk1,mrk2,lst
    integer, intent(inout) :: lst_lth
    real(kind=dp), intent(in) :: cutoff
    !-------------------------------------------------------------------------------
    ! Local Variables
    !
    integer, dimension(:), allocatable :: advance
    real(kind=dp), dimension(dimn) :: rij
    real(kind=dp) :: rsqij,rangesq
    integer :: i,j,L

    write(string,'(a)') 'Constructing neighbor list'
    allocate( advance(nat) )
    rangesq = cutoff**2
    L = 1
    do i = 1,nat
        do j = i+1,nat
            rij = at(i)%x - at(j)%x
            where (abs(rij)/samplesize > 0.5_dp .and. pbc)
                rij = rij-sign(samplesize,rij)
            end where
            rsqij = dot_product(rij,rij)
            if (rsqij < rangesq) then
                advance(j) = 1
            else
                advance(j) = 0
            endif
        enddo
        mrk1(i) = L
        do j = i+1,nat
            if ( L > size(lst)) then
                write(string,'(a)') "Neighbor List: List too small. Increase"
                call output
                write(string,'(a)') "the parameter max_pairs_per_atom in"
                call output
                write(string,'(a)') "module mod_interatomic. Stopped at atom "
                call output
                write(string,'(2I10)') i,mrk2(1)-mrk1(1)
                call output
                stop
            endif
            lst(L) = j
            L = L + advance(j)
        enddo
        mrk2(i) = L - 1
    enddo
    lst_lth = L - 1
    deallocate( advance )
    end subroutine ngh_list

    subroutine deallocate_neigh
    implicit none
    if (allocated(marker1  ))deallocate(   marker1  )
    if (allocated(marker2  ))deallocate(   marker2  )
    if (allocated(list     ))deallocate(     list   )
    end subroutine deallocate_neigh

end module mod_neigh


