!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2015, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!****************************************************************************
!**
!**   MODULE mod_particles: Particle properties
!**
!**   type definitions:
!**   ----------------
!**   N/A
!**
!**   Real Variable Definitions:
!**   -------------------------------------
!**   N/A
!**
!**   Type Variable Definitions:
!**   -------------------------------------
!**   type(atom) :: at(:)
!**
!**   Integer Variable Definitions:
!**   ----------------------------
!**   N/A
!**  
!**   logical Variable Definitions:
!**   ----------------------------
!**   N/A
!**
!**   Contains Routines:
!**   ------------------
!**   allocate_atom : Allocates memory to the type(atom) variable at
!**   read_config   : Reads in the positions, velocities and species of atoms
!**   deallocate_atoms : Deallocates atoms
!****************************************************************************
module mod_particles
use mod_global, only : dp
use mod_types, only : atom,species
implicit none
save
private
type(atom),allocatable, dimension(:) :: at,at_i,at_f
public allocate_atom,   &
       read_config,     &
       deallocate_atoms
public  at,at_i,at_f
contains
!---------------------------------------------------------------------------
!   allocate_atom : allocates memory to atoms 
!
!   Passed Parameters :-
!              dimn (in) : dimension
!              nat  (in) : number of atoms
!
    subroutine allocate_atom(dimn,nat)
    implicit none
    integer, intent(in) :: dimn,nat
    integer :: j
    allocate(at(nat))
    allocate(at_i(nat))
    allocate(at_f(nat))
    do j=1,nat
        allocate(at(j)%x(dimn))
        allocate(at(j)%v(dimn))
        allocate(at(j)%sp_ptr)

        allocate(at_i(j)%x(dimn))
        allocate(at_i(j)%v(dimn))
        allocate(at_i(j)%sp_ptr)

        allocate(at_f(j)%x(dimn))
        allocate(at_f(j)%v(dimn))
    enddo
    end subroutine allocate_atom
!---------------------------------------------------------------------------
!   read_config :  Reads in the positions, velocities and species 
!
!   Non-Obvious Parameters :-
!
    subroutine read_config(cfile)
    use mod_box, only : dimn,nat,nsp,max_nsp,spec,nsp_file,sp_list
    use mod_file, only : FileExists,iofile
    use mod_output, only : string,output, nerror,error,erroroutput
    implicit none
    character(len=80),intent(in) :: cfile

    character(len=160) fomt
    character(len=2) :: id
    integer :: ierror,i=0,iunit,j,k
    logical :: species_match,species_already_exists

    call allocate_atom(dimn,nat)

    if (FileExists(cfile,.false.)) then
        call iofile(cfile,'formatted  ',iunit,.false.,.false.)
    else   
        nerror = 2
        error(1) = 'configuration information file not found'
        error(2) = 'in subroutine read_config,  Filename = '& 
                    // trim(cfile)
        call erroroutput
        stop
     endif
        
    nsp=0
    do 
        i = i + 1
        if (i .le. nat) then
            at_i(i)%v = 0._dp
            read(iunit,*,iostat=ierror) id,at_i(i)%x
            if (ierror .ne. 0) then
                nerror=1
                write(error(1),'(a)') 'Premature end-of-file, Filename = '&
                                      // trim(cfile)
                call erroroutput
                stop
            endif

            ! Match the atomic species with the species id's in the species file
            species_match = .false.
            do j=1,nsp_file
                if (spec(j)%id .eq. id) then
                    at_i(i)%sp_ptr => spec(j)
                    species_match = .true.

                    ! See if the current species is different from the previous atomic species
                    ! If 'yes', then nsp = nsp+1
                    if (nsp .eq. 0) then
                        nsp=nsp+1
                        sp_list(nsp) = spec(j)%id
                    else
                        species_already_exists=.false.
                        do k=1,nsp
                            if (spec(j)%id .eq. sp_list(k)) then
                                species_already_exists = .true.
                                exit
                            endif
                        enddo
                        if (species_already_exists .eqv. .false.) then
                            nsp = nsp+1
                            sp_list(nsp) = spec(j)%id
                        endif
                    endif

                endif
            end do
            if (species_match .eqv. .false.) then
                nerror = 1
                write(error(1),1003) i,id
                call erroroutput
                stop
            endif

        elseif (i .gt. nat .and. i .le. 2*nat) then
            at_f(i-nat)%v = 0._dp
            !read(iunit,*,iostat=ierror) id,at(i)%x,at(i)%v
            ! Neglect id since we have already read the species from initial configuration
            read(iunit,*,iostat=ierror) id,at_f(i-nat)%x
            if (ierror .ne. 0) then
                nerror=1
                write(error(1),'(a)') 'Premature end-of-file, Filename = '&
                                      // trim(cfile)
                call erroroutput
                stop
            endif
        elseif (i .gt. 2*nat) then
            read(iunit,*,iostat=ierror)
            if (ierror .eq. 0) then
                nerror=3
                write(error(1),'(a)') 'More atoms than specified'
                write(error(2),1000) i
                write(error(3),1001) nat
                call erroroutput
                stop
            else
                exit
            endif
        endif
    end do
    do i=1,nat
        ! The default atomic positions are taken to be final coordinates
        ! This is changed to initial coordinates for calculating PK stress in mod_stress.f90
        at(i)%x = at_f(i)%x
        at(i)%sp_ptr = at_i(i)%sp_ptr
    enddo

    write(string,1004) nsp; call output
    write(fomt,'(a,I3,a)') '(',nsp,'a)'
    write(string,fomt) sp_list(1:nsp)
    !write(string,1005) sp_list(1:nsp); call output

    1000 format('i   = ',I10)
    1001 format('nat = ',I10)
    1003 format('Species of atom', I10,' = ',A2,' does not have a match in the species file')
    1004 format('Total number of species in the configuration file = ',I5)
    !1005 format('Species list : ',a)
    end subroutine read_config
!---------------------------------------------------------------------------
!   deallocate_atoms :  Deallocates atoms
!
!   Non-Obvious Parameters :-
!
    subroutine deallocate_atoms
    implicit none
    deallocate(at)
    end subroutine deallocate_atoms

end module mod_particles
