!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2015, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!-------------------------------------------------------------------------------
!** QCLIB:  These are a lot of handy utilities and other routines that
!**   don't fit neatly elsewhere in the code.  Most are self
!**   explanatory.  Here is a brief table of contents, but note that the
!**   actual order of the routines is different.
!**
!** vector and matrix manipulations:
!**    kronecker_delta : Kronecker delta
!**    balanc          : balance a matrix
!**    elmhes          : reduce matrix to Hessenberg form
!**    hqr             : get eigenvalues of a matrix
!**    cross_product   : 3-D cross product
!**    invmat33        : inverse of a 3x3 matrix
!**    detmat33        : determinant of a 3x3 matrix
!**    qsortr          : sort a list
!** geometric routines:
!**    PointInPoly     : check if a point is in a polygon
!**    between0        : check if the point zero is between two points
!**    betweentol      : check if the point zero is between two points 
!**                      to within a small tolerance
!**    between0_tol    : check if the point zero is between two points 
!**                      to within a small tolerance. (Slightly different
!**                      implementation than betweentol. Kept for downward
!**                      compatibility).
!**    intri           : check if a point is in a triangle
!**    QuadArea        : returns the area of a quadrilateral
!**    PolygonArea     : returns the area of a polygon
!**    RIntersectRatio : Given two points and a third point, return the
!**                      intersection ratio of the perpendicular
!**                      line form the third point to the line formed
!**                      by the first two points
!**    FindNearestPoint: Find the nearest point in the polygon to given point.
!** character string manipulations:
!**    freein          : parse free-form input lines
!**    downcase        : make a string all lower case
!**    next            : find the next delimiter (comma or space) in a string
!** QC specific, but called in many places:
!**    MapToRef        : map a point from deformed to reference configuration
!**    SMapToCurrent   : map a point from reference to deformed configuration
!**                      (shape-fn coordinates)
!**    MapToCurrent    : map a point from reference to deformed configuration
!**                      (real coordinates)
!**    MapToCurrent    : map a point from reference to deformed configuration
!**    RinfEigen       : find (rcut)*(max. eigenvalue of F)
!**    FEigenValues    : Compute EigenValues of F^T*F
!**    defg            :  Compute deformation gradient
!**    strain          :  Compute lagrangian strain
!** miscellaneous routines
!**    ran1            : random number generator (Numerical recipes)
!**    LoadLocal       : Load a local array for an element from the global data
!************************************************************************

module mod_qclib

public

contains

subroutine cross_product(x1,x2,x3)
use mod_global, only : dp
implicit none

!-- Transferred variables
real(kind=dp), intent(in)  :: x1(3),x2(3)
real(kind=dp), intent(out) :: x3(3)

x3(1)=x1(2)*x2(3) - x1(3)*x2(2)
x3(2)=x1(3)*x2(1) - x1(1)*x2(3)
x3(3)=x1(1)*x2(2) - x1(2)*x2(1)
return
end subroutine cross_product

subroutine tensor_prod(dimn,vec1,vec2,tensor)
use mod_global, only : dp
implicit none
integer, intent(in) :: dimn
real(kind=dp),intent(in) :: vec1(dimn),vec2(dimn)
real(kind=dp),intent(out):: tensor(dimn,dimn)
integer :: i,j
do i=1,dimn
    do j=1,dimn
        tensor(j,i) = vec1(j)*vec2(i)
    enddo
enddo
end subroutine tensor_prod

!** CHARACTER STRING MANIPULATIONS **!

!************************************************************************
integer function next(last,input)
implicit none
 
!-- Transferred variables
character(len=80) input
integer last
 
!-- Local Variables
integer i

do i = last+1,80
   if ((input(i:i).eq.',').or.(input(i:i).eq.' ')) then
!   if (input(i:i).eq.',') then
      next = i
      return
   end if
enddo
next = 81

return
end function next

!************************************************************************   
function downcase(s1)
implicit none

!-- Transferred variables
character(len=*), intent(in)  :: s1
character(len=len(s1)) downcase

!-- Local variables
integer i,ic

do i=1,len(s1)
   ic=ichar(s1(i:i))
   if ((ic > 64).and.(ic < 91)) then
      downcase(i:i)=achar(ic+32)
   else
      downcase(i:i)=s1(i:i)
   endif
enddo
end function downcase
!************************************************************************   
subroutine freein(input,ll,lr,intvar,realvar,isw)
use mod_global, only : dp
implicit none

!-- Transferred variables
integer ll,lr,intvar,isw
real(kind=dp) realvar
character(len=80) input
 
!-- Local Variables
character(len=1) range1
character(len=2) range2
character(len=80) strinp
integer ldif
 
strinp = input((ll+1):(lr-1))
ldif = lr - ll - 1
if (ldif.eq.0) then
   if (isw.eq.1) intvar = 0
   if (isw.eq.2) realvar = 0.0_dp
else
   if (ldif.lt.10) then
      write(range1,'(i1)') ldif
      if (isw.eq.1) read(strinp,'(i'//range1//')') intvar
      if (isw.eq.2) read(strinp,'(f'//range1//'.0)') realvar
   else
      write(range2,'(i2)') ldif
      if (isw.eq.1) read(strinp,'(i'//range2//')') intvar
      if (isw.eq.2) read(strinp,'(f'//range2//'.0)') realvar
   end if
end if
return
end subroutine freein

    subroutine intersect(arr1,arr2,arr12)
    implicit none

    integer, dimension(:),intent(in) :: arr1,arr2
    integer, dimension(:),intent(inout) :: arr12

    integer, dimension(size(arr1)) :: larr1
    integer, dimension(size(arr2)) :: larr2
    integer i,j,k,n

    larr1=arr1
    larr2=arr2

    call QsortC(larr1)
    call QsortC(larr2)

    k=1
    n=0
    arr12 = 0
    do i =1,size(larr1)
        do j=k,size(larr2)
            if (larr2(j) .le. larr1(i)) then
                if (larr2(j) .eq. larr1(i)) then
                    n=n+1
                    arr12(n) = larr1(i)
                    k=j+1
                    exit
                endif
            else
                exit
            endif
        enddo
    enddo
    end subroutine intersect

    recursive subroutine QsortC(A)
    integer, intent(in out), dimension(:) :: A
    integer :: iq

    if(size(A) > 1) then
        call Partition(A, iq)
        call QsortC(A(:iq-1))
        call QsortC(A(iq:))
    endif
    end subroutine QsortC

    subroutine Partition(A, marker)
    integer, intent(in out), dimension(:) :: A
    integer, intent(out) :: marker
    integer :: i, j
    integer:: temp
    integer:: x      ! pivot point
    x = A(1)
    i= 0
    j= size(A) + 1

    do
        j = j-1
        do
            if (A(j) <= x) exit
            j = j-1
        end do
        i = i+1
        do
            if (A(i) >= x) exit
            i = i+1
        end do
        if (i < j) then
            ! exchange A(i) and A(j)
            temp = A(i)
            A(i) = A(j)
            A(j) = temp
        elseif (i == j) then
            marker = i+1
            return
        else
            marker = i
            return
        endif
    end do
    end subroutine Partition

    subroutine amux ( n, x, y, a, ja, ia )
    !-------------------------------------------------------------------------------
    ! AMUX multiplies a CSR matrix A times a vector.
    !
    !  Discussion:
    !
    !    This routine multiplies a matrix by a vector using the dot product form.
    !    Matrix A is stored in compressed sparse row storage.
    !
    !  Modified:
    !
    !    07 January 2004
    !
    !  Author:
    !
    !    Youcef Saad
    !
    !  Parameters:
    !
    !    Input, integer N, the row dimension of the matrix.
    !
    !    Input, real X(*), and array of length equal to the column dimension
    !    of A.
    !
    !    Input, real A(*), integer JA(*), IA(NROW+1), the matrix in CSR
    !    Compressed Sparse Row format.
    !
    !    Output, real Y(N), the product A * X.
    !
    use mod_global, only  : dp
    implicit none

    integer n

    real ( kind = dp ) a(*)
    integer i
    integer ia(*)
    integer ja(*)
    integer k
    real ( kind = dp ) t
    real ( kind = dp ) x(*)
    real ( kind = dp ) y(n)

    do i = 1, n
    !
    !  Compute the inner product of row I with vector X.
    !
    t = 0.0D+00
    do k = ia(i), ia(i+1)-1
    t = t + a(k) * x(ja(k))
    end do

    y(i) = t

    end do

    return
    end subroutine amux

subroutine atmux ( m ,n, x, y, a, ja, ia )
!-------------------------------------------------------------------------------
!  ATMUX computes A' * x for a CSR matrix A.
!
!  Discussion:
!
!    This routine multiplies the transpose of a matrix by a vector when the
!    original matrix is stored in compressed sparse row storage. Can also be
!    viewed as the product of a matrix by a vector when the original
!    matrix is stored in the compressed sparse column format.
!
!  Modified:
!
!    07 January 2004
!
!  Author:
!
!    Youcef Saad
!
!  Parameters:
!
!    Input, integer N, the row dimension of the matrix A'.
!
!    Input, real X(*), an array whose length is equal to the
!    column dimension of A'.
!
!    Output, real Y(N), the product A' * X.
!
!    Input, real A(*), integer JA(*), IA(*), the matrix in CSR
!    Compressed Sparse Row format.
!
  use mod_global, only : dp

  implicit none

  integer m,n

  real ( kind = dp ) a(*)
  integer i
  integer ia(*)
  integer ja(*)
  integer k
  real ( kind = dp ) x(*)
  real ( kind = dp ) y(n)

  y = 0._dp

  do i = 1, m
    do k = ia(i), ia(i+1)-1
      y(ja(k)) = y(ja(k)) + x(i) * a(k)
    enddo
  enddo

  return
end subroutine atmux

  subroutine csrcsc ( n, job, ipos, a, ja, ia, ao, jao, iao )  
!-------------------------------------------------------------------------------
!  CSRCSC converts Compressed Sparse Row to Compressed Sparse Column.
!
!  Discussion:
!
!    This is essentially a transposition operation.
!
!    It is NOT an in-place algorithm.
!
!    This routine transposes a matrix stored in a, ja, ia format.
!
!  Modified:
!
!    07 January 2004
!
!  Author:
!
!    Youcef Saad
!
!  Parameters:
!
!    Input, integer N, the order of the matrix.
!
!    Input, integer JOB, indicates whether or not to fill the values of the
!    matrix AO or only the pattern (IA, and JA).  Enter 1 for yes.
!
! ipos  = starting position in ao, jao of the transposed matrix.
!         the iao array takes this into account (thus iao(1) is set to ipos.)
!         Note: this may be useful if one needs to append the data structure
!         of the transpose to that of A. In this case use
!                call csrcsc (n,1,n+2,a,ja,ia,a,ja,ia(n+2))
!        for any other normal usage, enter ipos=1.
!
!    Input, real A(*), integer JA(*), IA(N+1), the matrix in CSR
!    Compressed Sparse Row format.
!
!    Output, real AO(*), JAO(*), IAO(N+1), the matrix in CSC
!    Compressed Sparse Column format.
!
  use mod_global, only  : dp
  implicit none

  integer n

  real ( kind = dp ) a(*)
  real ( kind = dp ) ao(*)
  integer i
  integer ia(n+1)
  integer iao(n+1)
  integer ipos
  integer j
  integer ja(*)
  integer jao(*)
  integer job
  integer k
  integer next
!
!  Compute lengths of rows of A'.
!
  iao(1:n+1) = 0

  do i = 1, n
    do k = ia(i), ia(i+1)-1
      j = ja(k) + 1
      iao(j) = iao(j) + 1
    end do
  end do
!
!  Compute pointers from lengths.
!
  iao(1) = ipos
  do i = 1, n
    iao(i+1) = iao(i) + iao(i+1)
  end do
!
!  Do the actual copying.
!
  do i = 1, n
    do k = ia(i), ia(i+1)-1
      j = ja(k)
      next = iao(j)
      if ( job == 1 ) then
        ao(next) = a(k)
      end if
      jao(next) = i
      iao(j) = next + 1
    end do
  end do
!
!  Reshift IAO and leave.
!
  do i = n, 1, -1
    iao(i+1) = iao(i)
  end do
  iao(1) = ipos

  return
end subroutine csrcsc

end module mod_qclib
