!
! CDDL HEADER START
!
! The contents of this file are subject to the terms of the Common Development
! and Distribution License Version 1.0 (the "License").
!
! You can obtain a copy of the license at
! http://www.opensource.org/licenses/CDDL-1.0.  See the License for the
! specific language governing permissions and limitations under the License.
!
! When distributing Covered Code, include this CDDL HEADER in each file and
! include the License file in a prominent location with the name LICENSE.CDDL.
! If applicable, add the following below this CDDL HEADER, with the fields
! enclosed by brackets "[]" replaced with your own identifying information:
!
! Portions Copyright (c) [yyyy] [name of copyright owner]. All rights reserved.
!
! CDDL HEADER END
!
! Copyright (c) 2015, Regents of the University of Minnesota.
! All rights reserved.
!
! Contributors:
!    Nikhil C. Admal
!    Ellad B. Tadmor
!
! Release: This file is part of the MDStressLab git repository.
!
!****************************************************************************
!**
!**   MODULE mod_stress : Related to the stress tensor
!**
!**   type definitions:
!**   ----------------
!**   N/A
!**
!**   Real Variable Definitions:
!**   -------------------------------------
!**   avgsize : Size of the averaging domain
!**
!**   Type Variable Definitions:
!**   -------------------------------------
!**   type(atom) :: at(:)
!**
!**   Integer Variable Definitions:
!**   ----------------------------
!**   N/A
!**  
!**   logical Variable Definitions:
!**   ----------------------------
!**   flv : virial stress flag
!**   flt : tsai stress flag
!**   flh : hardy stress flag
!**
!**   Contains Routines:
!**   ------------------
!**   set_stress : Reads in stress related parameters
!**   calculate_stress : 
!****************************************************************************
module mod_stress
use mod_global, only : dp
implicit none
save
private
public set_stress,      &
       compute_stress,  &
       avgsize,         &
       write_stress

logical :: flv,flt,flh,pk
real(kind=dp) :: avgsize

contains

    subroutine set_stress
    use mod_global, only : cmdunit
    use mod_output, only : string,output,erroroutput,nerror,error,heading
    use mod_qclib, only : next,downcase,freein
    use mod_hardy, only : avgsize_hardy
    implicit none
    !-------------------------------------------------------------------------------
    ! local variables
    !
    character(len=80) :: input
    character(len=4) :: cc
    integer :: lower,upper,i,ldif,idum
    logical :: setflag,flag,terminate
    real(kind=dp) :: readin
    !-------------------------------------------------------------------------------
    ! allow indentation with spaces in input file
    !
    heading=.true.
    write(string,'(a)')'**Reading stress data**'; call output
    terminate=.false.
    do
        setflag=.false.
        read(cmdunit,'(a80)') input
        i=1
        do while ((input(i:i).eq.' ').and.(i.le.80))
            i=i+1
        enddo
        upper = i-1
        
        lower = upper
        upper = next(lower,input)
        ldif = upper - lower - 1
        if (ldif.eq.0) then
            cc = '    '
        else
            cc = input(lower+1:upper-1)
        end if
        
        lower=upper
        upper=next(lower,input)
        do i=lower,upper
             if ((input(i:i).eq.'T').or.(input(i:i).eq.'t')) then
                flag=.true.
                setflag=.true.
                exit
             else if ((input(i:i).eq.'F').or.(input(i:i).eq.'f')) then
                flag=.false.
                setflag=.true.
                exit
             endif
        enddo
        
        select case(downcase(cc))
        case('pkst')
            if (setflag .eqv. .false.) then
                nerror = 1
                error(1) = 'Unrecognized flag in the piola kirchhoff stress**'
                call erroroutput
                stop
            endif
            pk = flag
        case ('avgs')
            call freein(input,lower,upper,idum,readin,2)
            avgsize = readin
            avgsize_hardy = avgsize
        case('viri')
            if (setflag .eqv. .false.) then
                nerror = 1
                error(1) = 'Unrecognized flag in the virial stress**'
                call erroroutput
                stop
            endif
            flv = flag
        case('tsai')
            if (setflag .eqv. .false.) then
                nerror = 1
                error(1) = 'Unrecognized flag in the tsai stress**'
                call erroroutput
                stop
            endif
            flt = flag
        case('hard')
            if (setflag .eqv. .false.) then
                nerror = 1
                error(1) = 'Unrecognized flag in the hardy stress**'
                call erroroutput
                stop
            endif
            flh = flag
        case('end')
            terminate=.true.
        case default
            if (cc .ne.'    ' .and. cc .ne.'%') then
                nerror = 1
                error(1) = 'macro "' // cc // '" not recognized**'
                call erroroutput
                stop
            endif
        end select
        
        if (terminate) exit
    enddo

    write(string,5010) avgsize; call output
    if (pk) then
        write(string,5050) pk; call output
    endif
    if (flv) then
        write(string,5020) flv; call output
    endif
    if (flt) then 
        write(string,5030) flt; call output
    endif
    if (flh) then
        write(string,5040) flh; call output
    endif

    5010 format('Size of the averaging domain = ',f10.5)
    5050 format('calculate pk stress      = ',l5)
    5020 format('calculate virial stress      = ',l5)
    5030 format('calculate Tsai stress        = ',l5)
    5040 format('calculate Hardy stress       = ',l5)
    end subroutine set_stress

    subroutine compute_stress
    use mod_forces, only : compute_forces,rcutoff
    use mod_fij, only : fij
    use mod_output, only : string,output,heading
    use mod_box, only : dimn,nat,samplesize,samplesize_i
    use mod_particles, only : at,at_i
    use mod_neigh,only : list,marker1,marker2
    use mod_grid, only : ngrid,set_grid
    use mod_hardy,only : hardy,hardy_p,     &
                         virial,virial_p,   &
                         hardy_kin,  &
                         virial_kin, &
                         collect_hardy
    use mod_tsai,only  : tsai,tsai_p,collect_tsai,tsai_length,tsai_area
    use mod_rigidity, only : project,proj_fij
    implicit none

    integer :: i,l
    real(kind=dp) :: tensor(dimn,dimn),tensork(dimn,dimn)

    heading=.true.
    write(string,'(a)')'**Calculating the stress tensor**'; call output

    if (flh .or. flt .or. flv) then
        call compute_forces
    endif

    !-------------------------------------------------------------------------------
    ! If P--K stess is calculated then consider the reference configuration
    !
    if (pk) then
        do i=1,nat
            at(i)%x = at_i(i)%x
        enddo
        samplesize = samplesize_i
    endif

    !-------------------------------------------------------------------------------
    ! Call set_grid to setup the grid on either the reference (pk .eqv. T), or the present
    ! configuration (pk .eqv. F)
    !
    call set_grid(avgsize)


    if (flt) then
        tsai_length = 2._dp*avgsize
        tsai_area = tsai_length ** 2
        allocate(tsai(dimn,dimn,ngrid))
        tsai = 0._dp
        if (project) then 
            allocate(tsai_p(dimn,dimn,ngrid))
            tsai_p = 0._dp
        endif
    endif

    if (flh) then
        allocate(hardy(dimn,dimn,ngrid))
        allocate(hardy_kin(dimn,dimn,ngrid))
        hardy   = 0._dp
        hardy_kin   = 0._dp
        if (project) then
            allocate(hardy_p(dimn,dimn,ngrid))
            hardy_p = 0._dp
        endif
    endif
        
    if (flv) then
        allocate(virial(dimn,dimn,ngrid))
        allocate(virial_kin(dimn,dimn,ngrid))
        virial = 0._dp
        virial_kin = 0._dp
        if (project) then
            allocate(virial_p(dimn,dimn,ngrid))
            virial_p = 0._dp
        endif
    endif

    write(string,'(a)')' '; call output
    write(string,'(a)')'Looping over all bonds to calculate stress tensors'
    call output

    do l=1,ngrid
        if (real(l)/100 - int(real(l)/100) .eq. 0 ) write(*,*) "grid point: ",l
        if (flh) then
            tensor = 0._dp
            tensork = 0._dp
            call collect_hardy(l,.true.,rcutoff,fij,list,marker1,marker2,tensor,tensork)
            hardy(:,:,l) = hardy(:,:,l) + tensor
            hardy_kin(:,:,l) = hardy_kin(:,:,l) + tensork

            if (project) then
                tensor = 0._dp
                tensork = 0._dp
                call collect_hardy(l,.true.,rcutoff,proj_fij,list,marker1,marker2,tensor,tensork)
                hardy_p(:,:,l) = hardy_p(:,:,l) + tensor
            endif
        endif
        if (flt) then
            tensor = 0._dp
            call collect_tsai(l,rcutoff,fij,list,marker1,marker2,tensor)
            tsai(:,:,l) = tsai(:,:,l) + tensor
            if (project) then
                tensor = 0._dp
                call collect_tsai(l,rcutoff,proj_fij,list,marker1,marker2,tensor)
                tsai_p(:,:,l) = tsai_p(:,:,l) + tensor
            endif
        endif
       if (flv) then
            tensor = 0._dp
            tensork = 0._dp
            call collect_hardy(l,.false.,rcutoff,fij,list,marker1,marker2,tensor,tensork)
            virial(:,:,l) = virial(:,:,l) + tensor
            virial_kin(:,:,l) = virial_kin(:,:,l) + tensork

            if (project) then
                tensor = 0._dp
                tensork = 0._dp
                call collect_hardy(l,.false.,rcutoff,proj_fij,list,marker1,marker2,tensor,tensork)
                virial_p(:,:,l)   = virial_p(:,:,l) + tensor
            endif
       endif
    enddo

    end subroutine compute_stress

    subroutine write_stress
    use mod_hardy, only : write_hardy
    use mod_tsai,  only : write_tsai
    use mod_output, only : output,string
    implicit none
    write(string,'(a)')' '; call output
    if (flh) call write_hardy(1)
    if (flv) call write_hardy(0)
    if (flt) call write_tsai
    end subroutine write_stress

end module mod_stress
